/*
 * Copyright 1998 Vectaport Inc.
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <MapUnidraw/mapclasses.h>
#include <MapUnidraw/mapcmds.h>
#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapeditor.h>
#include <MapUnidraw/mapinspect.h>
#include <MapUnidraw/mapproj.h>
#include <MapUnidraw/mapstatecmds.h>
#include <MapUnidraw/mapstates.h>
#include <MapUnidraw/mapviews.h>
#include <MapUnidraw/mapviewer.h>
#include <MapUnidraw/mapviewstate.h>

#include <Map/mapcov.h>
#include <Map/mapdb.h>
#include <Map/mapelt.h>
#include <Map/mapfclass.h>
#include <Map/maplib.h>
#include <Map/mapprim.h>

#include <FrameUnidraw/frameclasses.h>

#include <OverlayUnidraw/ovdoer.h>
#include <OverlayUnidraw/ovviewer.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <UniIdraw/idvars.h>

#include <Unidraw/catalog.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/ctrlinfo.h>
#include <Unidraw/iterator.h>
#include <Unidraw/keymap.h>
#include <Unidraw/kybd.h>
#include <Unidraw/selection.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/Commands/edit.h>
#include <Unidraw/Graphic/ellipses.h>
#include <Unidraw/Graphic/lines.h>
#include <Unidraw/Graphic/polygons.h>

#include <InterViews/border.h>
#include <InterViews/box.h>
#include <InterViews/cursor.h>
#include <InterViews/frame.h>
#include <InterViews/glue.h>
#include <InterViews/menu.h>
#include <InterViews/message.h>
#include <InterViews/panner.h>
#include <InterViews/transformer.h>
#include <InterViews/tray.h>
#include <InterViews/window.h>

#include <Dispatch/rpcstream.h>

#include <ComTerp/comterpserv.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*****************************************************************************/

MapEditor::MapEditor(OverlayComp* gc, OverlayKit* ok) 
: FrameEditor(false, ok) {
    _projection = nil;
    _mapcomp = nil;
    _camcomp = nil;
    SetMapDatabase(nil);
    SetMapViewState(nil);

    Init(gc, "MapEditor");

    Initializing(true);
}

MapEditor::MapEditor(boolean initflag, OverlayKit* ok) 
: FrameEditor(initflag, ok) {
    SetMapViewState(nil);
    _mapcomp = nil;
}

StateVar* MapEditor::GetState (const char* name) {
    return IdrawEditor::GetState(name);
}

void MapEditor::Init (OverlayComp* comp, const char* name) {
    _curr_other = _prev_other = 0;
    _texteditor = nil;
    if (!comp) comp = new MapIdrawComp((const char*) nil, nil);
    _terp = new ComTerpServ();
    add_comterp("MapEditor", _terp);
    _overlay_kit->Init(comp, name);
    InitFrame();
}

void MapEditor::InitStateVars () {
    IdrawEditor::InitStateVars();
}

boolean MapEditor::Initializing() { return _init; }

void MapEditor::Initializing(boolean v) { _init = v; }

void MapEditor::OpenInit() {
    GetViewer()->Update();
    ((MapIdrawView*)GetViewer()->GetGraphicView())->SetMapEditor(this);

    dbstate()->notify();
    libstate()->notify();
    covstate()->notify();
    fclassstate()->notify();

    UpdateLibraryMenu();
    UpdateLayerMenu();
}

Interactor* MapEditor::Interior () {
    HBorder* hborder = new HBorder;
    VBorder* vborder = new VBorder;
    int gap = round(.1*ivcm);

    _tray->HBox(_tray, _viewer, _tray);

    _tray->VBox(_tray, _viewer, _tray);

    _tray->Align(BottomRight, _viewer, new Frame(new Panner(_viewer)));

    return _tray;
}

PJ* MapEditor::GetProjection() {
    return _projection;
}

void MapEditor::SetProjection(PJ* p) {
    _projection = p;
    GetMapViewState()->defproj(p);
}

void MapEditor::SetProjectionLocal(PJ* p) {
    _projection = p;
}

MapDatabase* MapEditor::GetMapDatabase() {
  return _mapdb;
}

void MapEditor::SetMapDatabase(MapDatabase* db) {
  _mapdb = db;
  if (_mapcomp) _mapcomp->SetDatabase(db);
}

MapDbComp* MapEditor::GetMapDbComp() {
    if (!_mapcomp)
	InitMapComp();
    return _mapcomp;
}

void MapEditor::SetMapDbComp(MapDbComp* dbcomp) {
    _mapcomp = dbcomp;
    if (dbcomp) 
	_mapdb = dbcomp->GetDatabase();
}

void MapEditor::InitMapComp() {
    _mapcomp = new MapDbComp(_mapdb);
    Clipboard* clb = new Clipboard();
    clb->Append(_mapcomp);
    PasteCmd* pcmd = new PasteCmd(this, clb);
    pcmd->Execute();
    Selection* s = GetViewer()->GetSelection();
    s->Clear();
}

MapDbView* MapEditor::GetMapDbView() {
    MapDbView* dbview = nil;
    GraphicView* views = GetViewer()->GetGraphicView();
    Iterator i;
    for (views->First(i); !views->Done(i); views->Next(i)) {
	GraphicView* view = views->GetView(i);
	if (view->IsA(MAPDB_VIEW)) {
	    dbview = (MapDbView*)view;
	    break;
	}
	else if (view->IsA(FRAME_VIEW)) {
	  Iterator j;
	  for (view->First(j); !view->Done(j); view->Next(j)) {
	    GraphicView* fview = view->GetView(j);
	    if (fview->IsA(MAPDB_VIEW)) {
	      dbview = (MapDbView*)fview;
	      return dbview;
	    }
	  }
	}
    }
    return dbview;
}

void MapEditor::Reproject() {
    PJ* p = GetProjection();
    GraphicComp* gc = (GraphicComp*)GetComponent();
    Iterator i;
    for (gc->First(i); !gc->Done(i); gc->Next(i)) {
      GraphicComp* comp = gc->GetComp(i);
      comp->Notify();
    }
}

void MapEditor::InitCommands() {
    if (_mapdb) {
        MapShowCmd* show = new MapShowCmd(this);
        show->SetMag(1./8.);
	CommandPusher* pusher = new CommandPusher(show);
	pusher->Push();
        _projstate->notify();
        _dbstate->notify();
        _libstate->notify();
        _covstate->notify();
        _fclassstate->notify();
    }
    FrameEditor::InitCommands();
}

void MapEditor::AddPoint(float lat, float lon) {
    MapPoint* point = new MapPoint();
    point->insert_points(1, &lon, &lat);
    MapFeature* feature = new MapFeature(nil, 0, MapFeature::PointType);
    feature->point_primitive(point);
    
    Clipboard* clb = new Clipboard();
    
    float mag = GetViewer()->GetMagnification();
    MapFeatureComp* mfc = new MapFeatureComp(feature);
    clb->Append(mfc);
    
    PasteCmd* pastecmd = new PasteCmd(this, clb);
    pastecmd->Execute();
    Selection* s = GetViewer()->GetSelection();
    s->Clear();

}

MapViewState* MapEditor::GetMapViewState() {
    if (!_mapview) 
        _mapview = new MapViewState();
    return _mapview;
}

void MapEditor::SetMapViewState(MapViewState* mv) {
    _mapview = mv;
}

void MapEditor::UpdateLayerMenu() {
    ((MapKit*)_overlay_kit)->UpdateLayerMenu();
}

void MapEditor::UpdateLibraryMenu() {
    ((MapKit*)_overlay_kit)->UpdateLibraryMenu();
}

boolean MapEditor::AdjustForGeo() {
  MapIdrawComp* comps = (MapIdrawComp*)GetComponent();
  Iterator i;
  comps->First(i);
  while (!comps->Done(i)) {
    OverlayComp* comp = (OverlayComp*)comps->GetComp(i);
    comps->Next(i);
    if (!comp) continue;
    AttributeList* al = ((OverlayComp*)comp)->attrlist();
    Attribute* scla;

#if 0
    if (al && (scla = al->GetAttr("scale"))) {
      float sv = scla->Value()->float_val();
      comp->GetGraphic()->Scale(sv, sv);
    }
#endif

    AttributeValue* sweav;
    AttributeValue* swnav;
    if ((sweav = comp->FindValue("sw_e")) 
	&& (swnav = comp->FindValue("sw_n"))) {
      double x1 = sweav->double_val();
      double y1 = swnav->double_val();
      float tx1, ty1;
      tx1 = x1; ty1 = y1;
      
      // use the mapviewstate mechanism
      MapViewState* vs = GetMapViewState();
      if (!vs->deftrans()) {
	// set it according to the first import coming in
        Transformer* tr = new Transformer();
	tr->translate(-tx1, -ty1);
	vs->settrans(tr);
      }
      if (!vs->implicitproj()) {
	AttributeValue* pav = comp->FindValue("proj");
	if (pav) {
	  char* pj = (char*)pav->string_ptr();
	  AttributeValue* elv = comp->FindValue("ellps");
	  char* el = (char*) (elv ? elv->string_ptr() : nil);
	  AttributeValue* znv = comp->FindValue("zone");
	  int zn = znv ? znv->int_val() : 0;
	  if (pj && el) {
	    char pjs[40];
	    char els[40];
	    char zs[40];
	    sprintf(pjs, "proj=%s", pj);
	    sprintf(els, "ellps=%s", el);
	    sprintf(zs, "zone=%d", zn);
	    char* parms[4] = { pjs, els, zs, "no_defs" };
	    PJ* iproj = pj_init(4, parms);
	    vs->implicitproj(iproj);
	    SetProjectionLocal(iproj);
	  }
	}
      }
    }
  }
}

void MapEditor::AddCommands(ComTerp* comterp) {
}
