/*
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Map/mapprim.h>
#include <Vpf/vpfcov.h>
#include <Vpf/vpffclass.h>
#include <Vpf/vpffeature.h>
#include <Vpf/vpflib.h>

#include <iostream.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

extern "C" {
#include <malloc.h>
#include <sys/time.h>
#include <sys/types.h>
}
#include <stdio.h>


/*****************************************************************************/

VpfFeatureClass::VpfFeatureClass (VpfCoverage* cov, const char* name, boolean makefeatures) 
: MapFeatureClass(cov, name) {
    _makefeatures = makefeatures;
    _tabl = NULL;
    _tdesc = NULL;
    _ptable = NULL;
    _desc = NULL;
    _features = NULL;
    _tilename = NULL;
    if (cov) {
      coverage(cov);
      VpfLibrary* lib = coverage()->library();
      VpfExtent* ext = lib->extent();
      _x1 = ext->x1;
      _y1 = ext->y1;
      _x2 = ext->x2;
      _y2 = ext->y2;
    }
    _zone = 0;
}

VpfFeatureClass::~VpfFeatureClass () {
    delete _tabl;
    delete _tdesc;
    delete _desc;
    delete_features();
    if (_columns) {
      for (int i = 0; i < _ncolumns; i++)
	delete _columns[i];
      delete [] _columns;
    }
}

void VpfFeatureClass::coverage(VpfCoverage* cov) { 
    _coverage = (MapCoverage*)cov; 
    if (_coverage) {
        featuretable();
        primitivetable();
        _isprim = is_primitive(_tabl);
        _issimpf = is_simple_feature(_tabl);
        _iscompf = is_complex_feature(_tabl);
        _isfeat = is_feature(_tabl);
        _isjoin = is_join(_tabl);
        _primclass = primitive_class(_tabl);
        if (_primclass == 0 && strcmp(_coverage->name(), "libref")==0)
	    _primclass = EDGE;
        _fctype = feature_class_type(_tabl);
        if (_fctype == 0 && strcmp(_coverage->name(), "libref")==0)
	    _fctype = LINE;
        description();
        if (_makefeatures) {
	    _features = make_features(_nfeats);
	if (!_features)
	    cerr << "Error making features\n";
        }
        make_columns();
    }
}

const char* VpfFeatureClass::description() const {
    if (!_desc) {
	((VpfFeatureClass*)this)->_desc = new char[80];
	char* desc = ((VpfFeatureClass*)this)->_desc;
	strcpy(desc, coverage()->description());
	switch (_fctype) {
	case LINE:
	    strcat(desc," Lines"); break;
	case AREA:
	    strcat(desc," Areas"); break;
	case ANNO:
	    strcat(desc," Text"); break;
	case POINT:
	    strcat(desc," Points"); break;
	case COMPLEX_FEATURE:
	    strcat(desc," Complex"); break;
	case 0:
	    strcat(desc," None"); break;
	default:
	    strcat(desc," Unknown"); break;
	}
    }
    return _desc;
}

void VpfFeatureClass::print() const {
    printf("   Feature Class: %s\n", _name);
    if (_desc) {
	printf("   Description: %s\n", _desc);
	printf("    Table: %s\n", _tabl);
	printf("    Table Description: %s\n", _tdesc);
	printf("    Primitive Table: %s\n", _ptable);
	printf("    IsPrimitive %d, IsSimpleFeature %d, IsComplexFeature %d\n",
	       _isprim, _issimpf, _iscompf);
	printf("    IsFeature %d, IsJoin %d\n", _isfeat, _isjoin);
	printf("    Primitive Class: %d ", _primclass);
	switch (_primclass) {
	case EDGE:
	    printf("(EDGE)\n"); break;
	case FACE:
	    printf("(FACE)\n"); break;
	case TEXT:
	    printf("(TEXT)\n"); break;
	case ENTITY_NODE:
	    printf("(ENTITY_NODE)\n"); break;
	case CONNECTED_NODE:
	    printf("(CONNECTED_NODE)\n"); break;
	case 0:
	    printf("(not primitive)\n"); break;
	default:
	    printf("(unknown value)\n"); break;
	}
	printf("    Feature Class Type: %d ", _fctype);
	switch (_fctype) {
	case LINE:
	    printf("(LINE)\n"); break;
	case AREA:
	    printf("(AREA)\n"); break;
	case ANNO:
	    printf("(ANNO)\n"); break;
	case POINT:
	    printf("(POINT)\n"); break;
	case COMPLEX_FEATURE:
	    printf("(COMPLEX_FEATURE)\n"); break;
	case 0:
	    printf("(none)\n"); break;
	default:
	    printf("(unknown value)\n"); break;
	}
	printf("Number of Columns: %d\n", ncolumns);
	for (int i = 1; i <= _ncolumns; i++) {
	    printf("Column number %d:\n", i);
	    _columns[i-1]->print();
	}
// Dump table in all its glory
	printf("    Feature Table Dump:\n");
	dump_table();
	printf("    Primitive Table Dump:\n");
	dump_prim_table();
	printf("    Ring Table Dump:\n");
	dump_ring_table();
    }
}

int VpfFeatureClass::is_primitive(const char* tname) {
    char namebuf[strlen(tname)+1];
    char* name = namebuf;
    strcpy(name, tname);
    char* end = strrchr(name, '/');
    if (end) {
	end += sizeof(char);
	name = end;
    }
    int len = strlen(name);
    if (name[len-1]=='.')
	name[len-1]='\0';
    
    if (strcasecmp(name, "end")==0 ||
	strcasecmp(name, "cnd")==0 ||
	strcasecmp(name, "edg")==0 ||
	strcasecmp(name, "fac")==0 ||
	strcasecmp(name, "txt")==0)
	return 1;
    else
	return 0;
}

int VpfFeatureClass::primitive_class(const char* tname) {
    char namebuf[strlen(tname)+1];
    char* name = namebuf;
    strcpy(name, tname);
    const char* end = strrchr(tname, '/');
    if (end) {
	end += sizeof(char);
	name = (char*) end;
    }
    int len = strlen(name);
    if (name[len-1]=='.')
	 name[len-1]='\0';
	 if (strcasecmp(name, "end")==0) return ENTITY_NODE;
	 if (strcasecmp(name, "cnd")==0) return CONNECTED_NODE;
	 if (strcasecmp(name, "edg")==0) return EDGE;
	 if (strcasecmp(name, "fac")==0) return FACE;
	 if (strcasecmp(name, "txt")==0) return TEXT;
	 return 0;
}

int VpfFeatureClass::is_complex_feature(const char* tname) {
    const char* suffix = strrchr(tname, '.');
    if (!suffix)
	return 0;
    if (strcasecmp(suffix, ".cft")==0)
	return 1;
    else
	return 0;
}

int VpfFeatureClass::is_simple_feature(const char* tname) {
    const char* suffix = strrchr(tname, '.');
    if (!suffix)
	return 0;
    if (strcasecmp(suffix, ".pft")==0 ||
	strcasecmp(suffix, ".lft")==0 ||
	strcasecmp(suffix, ".aft")==0 ||
	strcasecmp(suffix, ".tft")==0)
	return 1;
    else
	return 0;
}

int VpfFeatureClass::feature_class_type(const char* tname) {
    const char* suffix = strrchr(tname, '.');
    if (!suffix)
	return 0;
    if (strcasecmp(suffix, ".pft")==0) return POINT;
    if (strcasecmp(suffix, ".lft")==0) return LINE;
    if (strcasecmp(suffix, ".aft")==0) return AREA;
    if (strcasecmp(suffix, ".tft")==0) return ANNO;
    if (strcasecmp(suffix, ".cft")==0) return COMPLEX_FEATURE;
    return 0;
}

int VpfFeatureClass::is_feature(const char* tname) {
  return is_simple_feature(tname) || is_complex_feature(tname);
}

int VpfFeatureClass::is_join(const char* tname) {
    const char* suffix = strrchr(tname, '.');
    if (!suffix)
	return 0;
    if (strcasecmp(suffix, ".cjt")==0 ||
	strcasecmp(suffix, ".pjt")==0 ||
	strcasecmp(suffix, ".ljt")==0 ||
	strcasecmp(suffix, ".ajt")==0 ||
	strcasecmp(suffix, ".tjt")==0)
	return 1;
    else
	return 0;
}

void VpfFeatureClass::tilename(char* tn) {
  _tilename = tn;
  delete [] _ptable;
  _ptable = nil;
  primitivetable();
}

const char* VpfFeatureClass::primitivetable() const {
    if (!_ptable) {
	char path[255];
	vpf_table_type table;
	row_type row;
	long int i,n;
	int t1fieldnum, fcfieldnum, ffound, tfound;
	char *table1, *fc;
	struct stat statbuf;
	((VpfFeatureClass*)this)->_ptable = new char[255];
	char* ptable = ((VpfFeatureClass*)this)->_ptable;
	const char* covpath = _coverage->path();
	strcpy(ptable, covpath);
	strcat(ptable, "/fcs.");
	if (stat(ptable, &statbuf) != 0 &&
	    (ptable[strlen(ptable)-1]='\0', stat(ptable, &statbuf) != 0))
	{
	    if (strcmp(_coverage->name(), "libref")!=0) {
		cerr << "VpfFeatureClass::primitivetable: ";
		cerr << "Invalid VPF coverage (" << covpath << ")- missing FCS\n";
	    }
	    strcpy(ptable, "");
	}
	else {
	    table = vpf_open_table(ptable,disk,"rb",NULL);
	    if (!table.fp) {
		strcpy(ptable, "");
	    }
	    else {
		fcfieldnum = table_pos("FEATURE_CLASS",table);
		if (fcfieldnum < 0) {
		}
		else {
		    t1fieldnum = table_pos("TABLE2",table);
		    if (t1fieldnum < 0) {
		    }
		    else {
			ffound = tfound = 0;
			for (i = 1; i <= table.nrows; i++) {
			    row = read_next_row(table);
			    fc = (char *)get_table_element(fcfieldnum,row,table,NULL,&n);
			    rightjust(fc);
			    if (strcasecmp(fc, _name) == 0) {
				ffound = 1;
				table1 = (char *)get_table_element(t1fieldnum,row,table,NULL,&n);
				rightjust(table1);
				if (is_primitive(table1)) {
				    tfound = 0;
				    strcpy(ptable, covpath);
				    strcat(ptable, "/");
				    if (_tilename) {
				      strcat(ptable, _tilename);
				      strcat(ptable, "/");
				    }
				    strcat(ptable,strlower(table1));
				    strcat(ptable,".");
				}
				free(table1);
			    }
			    free_row(row,table);
			    free(fc);
			}
			//vpf_close_table(&table);
			
			if (!ffound) {
			}
			if (!tfound) {
			}
		    }
		}
		vpf_close_table(&table);
	    }
	}
    }
    return _ptable;
}

const char* VpfFeatureClass::featuretable() const {
    if (!_tabl) {
	char path[255];
	vpf_table_type table;
	row_type row;
	long int i,n;
	int t1fieldnum, fcfieldnum, ffound, tfound;
	char *table1, *fc;
	struct stat statbuf;
	((VpfFeatureClass*)this)->_tabl = new char[255];
	char *tabl = ((VpfFeatureClass*)this)->_tabl;
	const char* covpath = _coverage->path();
	strcpy(tabl, covpath);
	strcat(tabl, "/fcs.");
	if (stat(tabl, &statbuf) != 0 &&
	    (tabl[strlen(tabl)-1]='\0', stat(tabl, &statbuf) != 0))
	{
	    if (strcmp(_coverage->name(), "libref")!=0) {
		cerr << "VpfFeatureClass::featuretable: ";
		cerr << "Invalid VPF coverage (" << covpath << ") - missing FCS\n";
	    }
	    strcpy(tabl, "");
	}
	else {
	    table = vpf_open_table(tabl,disk,"rb",NULL);
	    if (!table.fp) {
		strcpy(tabl, "");
	    }
	    else {
		fcfieldnum = table_pos("FEATURE_CLASS",table);
		if (fcfieldnum < 0) {
		}
		else {
		    t1fieldnum = table_pos("TABLE1",table);
		    if (t1fieldnum < 0) {
		    }
		    else {
			ffound = tfound = 0;
			for (i = 1; i <= table.nrows; i++) {
			    row = read_next_row(table);
			    fc = (char *)get_table_element(fcfieldnum,row,table,NULL,&n);
			    rightjust(fc);
			    if (strcasecmp(fc, _name) == 0) {
				ffound = 1;
				table1 = (char *)get_table_element(t1fieldnum,row,table,NULL,&n);
				rightjust(table1);
				if (is_feature(table1)) {
				    tfound = 0;
				    if (is_feature(tabl)) {
					if (is_complex_feature(table1)) {
					    strcpy(tabl,covpath);
					    strcat(tabl,strlower(table1));
					}
				    }
				    else {
					strcpy(tabl, covpath);
					strcat(tabl, "/");
					strcat(tabl,strlower(table1));
				    }
				}
				free(table1);
			    }
			    free_row(row,table);
			    free(fc);
			}
			//vpf_close_table(&table);
			
			if (!ffound) {
			}
			if (!tfound) {
			}
		    }
		}
		vpf_close_table(&table);
	    }
	}
    }
    return _tabl;
}

int VpfFeatureClass::is_primitive() const {
    return _isprim;
}

int VpfFeatureClass::is_simple_feature() const {
    return _issimpf;
}

int VpfFeatureClass::is_complex_feature() const {
    return _iscompf;
}

int VpfFeatureClass::is_feature() const {
    return _isfeat;
}

int VpfFeatureClass::is_join() const {
    return _isjoin;
}

int VpfFeatureClass::primitive_class() const {
    return _primclass;
}

int VpfFeatureClass::feature_class_type() const {
    return _fctype;
}

void VpfFeatureClass::dump_table() const {
    if (_tabl)
	pretty_dump_table(_tabl, stdout);
}

void VpfFeatureClass::dump_prim_table() const {
    char cpath[255];
    strcpy(cpath, _coverage->path());
    strcat(cpath, "/");
    switch (_fctype) {
    case LINE:
	strcat(cpath, "edg."); break;
    case AREA:
	strcat(cpath, "fac."); break;
    default:
	return; break;
    }
    pretty_dump_table(cpath, stdout);
}

void VpfFeatureClass::dump_ring_table() const {
    char cpath[255];
    strcpy(cpath, _coverage->path());
    strcat(cpath, "/rng.");
    pretty_dump_table(cpath, stdout);
}

void VpfFeatureClass::dump_dict() const {
    char cpath[255];
    strcpy(cpath, _coverage->path());
    strcat(cpath, "/");
//  feature_class_dictionary(cpath, name, stdout);
}

void VpfFeatureClass::delete_features() {
    if (_features) {
	for (int i = 0; i < _nfeats; i++)
	    delete _features[i];
	delete [] _features;
	_features = NULL;
	_nfeats = 0;
    }
}

void VpfFeatureClass::make_columns() {
    struct stat buf;
    if (_tabl[0] && stat(_tabl, &buf) == 0) {
	vpf_table_type ftable;
	ftable = vpf_open_table(_tabl,disk,"rb",NULL); 
	_ncolumns = ftable.nfields;
	_columns = new VpfColumnHeader*[_ncolumns];
	for (int i = 0; i < _ncolumns; i++) {
	    _columns[i] = new VpfColumnHeader(ftable.header[i]);
	}
	vpf_close_table(&ftable);
    }
    else {
	_ncolumns = 0;
	_columns = NULL;
    }
}

MapFeature** VpfFeatureClass::make_features(int &nfeats) {
    MapFeature** features;
    if (_fctype == AREA) {
	vpf_table_type feattable, facetable, ringtable, edgetable;
	long int   i,k,n;
	char       *buf,  ch, date[40];
	row_type   featrow, facerow, ringrow, edgerow;
	char featpath[255], facepath[255], ringpath[255], edgepath[255];
	boolean notype = false;
	
	strcpy(featpath, _tabl);
	strcpy(facepath, _ptable);
	strcpy(ringpath, _coverage->path());
	strcpy(edgepath, ringpath);
	if (_tilename) {
	  strcat(ringpath, "/");
	  strcat(ringpath, _tilename);
	  strcat(edgepath, "/");
	  strcat(edgepath, _tilename);
	}
	strcat(ringpath, "/rng.");
	strcat(edgepath, "/edg.");
	
	feattable = vpf_open_table(featpath,disk,"rb",NULL);
	if (!feattable.fp) return false;
	long int feat_id_pos = table_pos("ID", feattable);
	long int feat_prim_id_pos = table_pos("FAC_ID", feattable);
	if (strcmp(name(),"tileref")==0)
	    notype = true;
	long int feat_type_pos;
	if (!notype) {
	    char tname[20];
	    const char* fcname = name();
	    tname[0] = toupper(fcname[0]);
	    tname[1] = toupper(fcname[1]);
	    tname[2] = '\0';
	    strcat(tname, "PYTYPE");
	    feat_type_pos = table_pos(tname, feattable);
	    if (feat_type_pos == -1) //return false;
		notype = true;
	}
	facetable = ((VpfCoverage*)_coverage)->facetable(_tilename);
	if (!facetable.fp) return false;
	long face_id_pos = table_pos("ID", facetable);
	long face_ringptr_pos = table_pos("RING_PTR", facetable);
	ringtable = vpf_open_table(ringpath,disk,"rb",NULL);
	if (!ringtable.fp) return false;
	long ring_id_pos = table_pos("ID", ringtable);
	long ring_faceid_pos = table_pos("FACE_ID", ringtable);
	long ring_startedge_pos = table_pos("START_EDGE", ringtable);
	edgetable = ((VpfCoverage*)_coverage)->edgetable(_tilename);
	if (!edgetable.fp) return false;
	long edge_id_pos = table_pos("ID", edgetable);
	long edge_rface_pos = table_pos("RIGHT_FACE", edgetable);
	long edge_lface_pos = table_pos("LEFT_FACE", edgetable);
	long edge_right_pos = table_pos("RIGHT_EDGE", edgetable);
	long edge_left_pos = table_pos("LEFT_EDGE", edgetable);
	long edge_coord_pos = table_pos("COORDINATES", edgetable);
	
	nfeats = feattable.nrows;
	features = (MapFeature**) new VpfFeature*[nfeats];
// this works only because there is a one-to-one correspondence of features
// to primitives in untiled feature classes!
	for (i=1;i<=feattable.nrows;i++) {
	    featrow = get_row(i,feattable);
	    long int featid, feattype;
	    get_table_element(feat_id_pos,featrow,feattable,&featid,&n);
	    if (!notype) {
		get_table_element(feat_type_pos,featrow,feattable,&feattype,&n);
		features[i-1] = outline_face(featrow,featid,featid,facetable,ringtable,edgetable,
					     feattype,0);
	    }
	    else {
	      if ((strcmp(name(),"tileref")==0
		   || strcmp(coverage()->library()->name(),"rference")==0)
		  && feat_prim_id_pos != -1) {
		long int prim_id;
		get_table_element(feat_prim_id_pos,featrow,feattable,&prim_id,
				  &n);
		features[i-1] = outline_face(featrow,featid,prim_id,
					     facetable,ringtable,edgetable,
					     0,0);
	      }
	      else
		features[i-1] = outline_face(featrow,featid,featid,facetable,ringtable,edgetable,
					     0,0);
	    }
	}
	vpf_close_table( &feattable );
	vpf_close_table( &ringtable );
	vpf_close_table( &facetable );
	vpf_close_table( &edgetable );
	return features;
    }
    else if (_fctype == LINE) {
	vpf_table_type feattable, edgetable;
	long int   i,k,n;
	char       *buf,  ch, date[40];
	row_type   featrow, edgerow;
	char featpath[255], edgepath[255];
	edge_rec_type edge_rec;
	struct stat statbuf;
	boolean notype = false;
	
	strcpy(featpath, _tabl);
	strcpy(edgepath, _coverage->path());
	strcat(edgepath, "/edg.");
	
	if (stat(featpath, &statbuf)==0)
	    feattable = vpf_open_table(featpath,disk,"rb",NULL);
	else
	    feattable.fp = NULL;
	if (!feattable.fp) {
	    if (strcmp(_coverage->name(),"libref")!=0)
		return false;
	    else { //need to handle libref here
		edgetable = ((VpfCoverage*)_coverage)->edgetable(_tilename);
		if (!edgetable.fp) return false;
		long edge_id_pos = table_pos("ID", edgetable);
		nfeats = edgetable.nrows;
		features = (MapFeature**) new VpfFeature*[nfeats];
		for (i=1;i<=edgetable.nrows;i++) {
		    edge_rec = read_edge(i, edgetable);
		    int n = edge_rec.npts;
		    float* x = new float[n];
		    float* y = new float[n];
		    coordinate_type coord;

		    for (int j = 0; j < n; j++) {
			coord = get_edge_coordinate(j,&edge_rec);
			x[j] = coord.x;
			y[j] = coord.y;
		    }
		    MapEdge* me = new MapEdge;
		    me->insert_pointers(n, x, y, nil, true);
		    features[i-1] = new VpfFeature((VpfFeatureClass*)this,
						   i,
						   MapFeature::LineType, 
						   0, nil);
		    features[i-1]->edge_primitive(me);
		}
		vpf_close_table( &edgetable );
		return features;
	    }
	}
	long int feat_id_pos = table_pos("ID", feattable);
	char tname[20];
	const char* fcname = name();
	tname[0] = toupper(fcname[0]);
	tname[1] = toupper(fcname[1]);
	tname[2] = '\0';
	strcat(tname, "LNTYPE");
	long int feat_type_pos = table_pos(tname, feattable);
	if (feat_type_pos == -1) //return false;
	    notype = true;
	edgetable = ((VpfCoverage*)_coverage)->edgetable(_tilename);
	if (!edgetable.fp) return false;
	long edge_id_pos = table_pos("ID", edgetable);
	long edge_rface_pos = table_pos("RIGHT_FACE", edgetable);
	long edge_lface_pos = table_pos("LEFT_FACE", edgetable);
	long edge_right_pos = table_pos("RIGHT_EDGE", edgetable);
	long edge_left_pos = table_pos("LEFT_EDGE", edgetable);
	long edge_coord_pos = table_pos("COORDINATES", edgetable);
	
	nfeats = feattable.nrows;
	features = (MapFeature**) new VpfFeature*[nfeats];
	for (i=1;i<=feattable.nrows;i++) {
	    featrow = get_row(i,feattable);
	    long int featid, feattype, featedge;
	    get_table_element(feat_id_pos,featrow,feattable,&featid,&n);
	    if (!notype)
		get_table_element(feat_type_pos,featrow,feattable,&feattype,&n);
	    else
		feattype = 0;
	    edge_rec = read_edge(featid, edgetable);
	    int n = edge_rec.npts;
	    float* x = new float[n];
	    float* y = new float[n];
	    coordinate_type coord;

	    for (int j = 0; j < n; j++) {
		coord = get_edge_coordinate(j,&edge_rec);
		x[j] = coord.x;
		y[j] = coord.y;
	    }
	    MapEdge* me = new MapEdge;
	    me->insert_pointers(n, x, y, nil, true);
	    features[i-1] = new VpfFeature((VpfFeatureClass*)this, featid, MapFeature::LineType, 
					    feattype, featrow);
	    features[i-1]->edge_primitive(me);
	}
	vpf_close_table( &feattable );
	vpf_close_table( &edgetable );
	return features;
    }
    else if (_fctype == POINT) {
	vpf_table_type feattable, nodetable;
	long int   i,j,k,n;
	char       *buf,  ch, date[40];
	row_type   featrow, noderow;
	char featpath[255], nodepath[255];
	point_rec_type node_rec;
	boolean notype = false;
	
	strcpy(featpath, _tabl);
	strcpy(nodepath, _coverage->path());
	strcat(nodepath, "/end.");
	
	feattable = vpf_open_table(featpath,disk,"rb",NULL);
	if (!feattable.fp) return false;
	long int feat_id_pos = table_pos("ID", feattable);
	char tname[20];
	const char* fcname = name();
	tname[0] = toupper(fcname[0]);
	tname[1] = toupper(fcname[1]);
	tname[2] = '\0';
	strcat(tname, "PTTYPE");
	long int feat_type_pos = table_pos(tname, feattable);
	if (feat_type_pos == -1) //return false;
	    notype = true;
	nodetable = vpf_open_table(nodepath,disk,"rb",NULL);
	if (!nodetable.fp) return false;
	long node_id_pos = table_pos("ID", nodetable);
	long node_coord_pos = table_pos("COORDINATE", nodetable);
	
	nfeats = feattable.nrows;
	features = (MapFeature**) new VpfFeature*[nfeats];
	for (i=1;i<=feattable.nrows;i++) {
	    featrow = get_row(i,feattable);
	    long int featid, feattype, featnode;
	    get_table_element(feat_id_pos,featrow,feattable,&featid,&n);
	    if (!notype)
		get_table_element(feat_type_pos,featrow,feattable,&feattype,&n);
	    else
		feattype = 0;
	    node_rec = read_point(featid, nodetable);
	    float dx = node_rec.x;
	    float dy = node_rec.y;
	    MapPoint* mp = new MapPoint();
	    mp->insert_points(1, &dx, &dy);
	    features[i-1] = new VpfFeature((VpfFeatureClass*)this, featid, MapFeature::PointType, 
					   feattype, featrow);
	    features[i-1]->point_primitive(mp);
	}
	vpf_close_table( &feattable );
	vpf_close_table( &nodetable );
	return features;
    }
    else if (_fctype == ANNO) {
	vpf_table_type feattable, texttable;
	long int   i,j,k,n;
	char       *buf,  ch, date[40];
	row_type   featrow, textrow;
	char featpath[255], textpath[255];
	anno_rec_type anno_rec;
	
	strcpy(featpath, _tabl);
	strcpy(textpath, _coverage->path());
	strcat(textpath, "/txt.");
	
	feattable = vpf_open_table(featpath,disk,"rb",NULL);
	if (!feattable.fp) return false;
	long int feat_id_pos = table_pos("ID", feattable);
	texttable = vpf_open_table(textpath,disk,"rb",NULL);
	if (!texttable.fp) return false;
	long text_id_pos = table_pos("ID", texttable);
	long text_coord_pos = table_pos("SHAPE_LINE", texttable);
	
	nfeats = feattable.nrows;
	features = (MapFeature**) new VpfFeature*[nfeats];
	for (i=1;i<=feattable.nrows;i++) {
	    featrow = get_row(i,feattable);
	    long int featid, feattype, featnode;
	    get_table_element(feat_id_pos,featrow,feattable,&featid,&n);
	    anno_rec = read_anno(featid, texttable);
	    float dx = anno_rec.x;
	    float dy = anno_rec.y;
	    MapText* mt = new MapText();
	    mt->text(anno_rec.text);
	    mt->insert_points(1, &dx, &dy);
	    features[i-1] = new VpfFeature((VpfFeatureClass*)this, featid, MapFeature::TextType, 
					   0, featrow);
	    features[i-1]->text_primitive(mt);
	}
	vpf_close_table( &feattable );
	vpf_close_table( &texttable );
	return features;
    }
    nfeats = 0;
    return NULL;
}

void VpfFeatureClass::add_polygon_edge(edge_rec_type *edge_rec,
				       float*& x, float*& y,
				       int* n, int* size) const {
  if (edge_rec->left == edge_rec->right) return;
  coordinate_type coord;
  if (*n+edge_rec->npts >= *size) {
    float* newx, *newy;
    *size += edge_rec->npts +1;
    newx = new float[*size];
    newy = new float[*size];
    for (int i = 0; i < *n; i++) {
      newx[i] = x[i];
      newy[i] = y[i];
    }
    delete [] x;
    delete [] y;
    x = newx;
    y = newy;
  }
  register long int i,start,end,inc,cnt;
  if (edge_rec->dir == '-') {
    start = edge_rec->npts-1;
    end = 0;
    inc = -1;
  }
  else {
    start = 0;
    end = edge_rec->npts-1;
    inc = 1;
  }
  for (i=start,cnt=0; i!=end; i+=inc,cnt++) {
    coord = get_edge_coordinate(i,edge_rec);
    x[(*n)+cnt] = coord.x;
    y[(*n)+cnt] = coord.y;
  }
  coord = get_edge_coordinate(i,edge_rec);
  x[(*n)+cnt] = coord.x;
  y[(*n)+cnt] = coord.y;
  (*n) += edge_rec->npts;
}

long int VpfFeatureClass::next_polygon_edge( edge_rec_type *edge_rec, long int *prevnode,
			    long int face_id ) const {
   long int next;

   if ( (edge_rec->right == face_id) && (edge_rec->left == face_id) ) {
      /* Dangle - go the opposite direction to continue along the boundary */
      if (*prevnode == edge_rec->start) {
	 edge_rec->dir = '-';
	 next = edge_rec->leftfwd;
	 *prevnode = edge_rec->start;
      } else if (*prevnode == edge_rec->end) {
	 edge_rec->dir = '+';
	 next = edge_rec->rightfwd;
	 *prevnode = edge_rec->end;
      } else next = -1;
   } else if (edge_rec->right == face_id) {
      /* The face is on the right - take the right forward edge */
      next = edge_rec->rightfwd;
      edge_rec->dir = '+';
      *prevnode = edge_rec->end;
   } else if (edge_rec->left == face_id) {
      /* The face is on the left - take the left forward edge */
      next = edge_rec->leftfwd;
      edge_rec->dir = '-';
      *prevnode = edge_rec->start;
   } else next = -1;

   return next;
}

MapFace* VpfFeatureClass::outline_polygon_loop(row_type row, long int face_id,
				     long int start_edge,
				     vpf_table_type edgetable, int ftype) const
{
    MapFace* mface = new MapFace();
    edge_rec_type edge_rec;
    long int next, prevnode;
    boolean done=FALSE;
    edge_rec = read_edge( start_edge, edgetable );
    edge_rec.dir = '+';
    prevnode = edge_rec.start;
    if (edge_rec.start != edge_rec.end) {
	next = next_polygon_edge( &edge_rec, &prevnode, face_id );
    } else {
	done = TRUE;
    }
    int size = 20;
    int n = 0;
    float* x = new float[size];
    float* y = new float[size];
    add_polygon_edge(&edge_rec, x, y, &n, &size);
    if (edge_rec.coord) free(edge_rec.coord);
    MapEdge* edge = new MapEdge;
    edge->insert_pointers(n, x, y, nil, true);
    MapNode* start_node, *end_node;
    while (!done) {
	if (next < 0) {
	    printf(" VPF topology error! Edge: %ld   face: %ld   left: %ld   right: %ld\n",
		   edge_rec.id,face_id,edge_rec.left,edge_rec.right);
	    done = TRUE;
	}
	if (next==0) {
	    done = TRUE;
	}
	if (next == start_edge) done = TRUE;
	if (!done) {
	    edge_rec = read_edge( next, edgetable );
	    next = next_polygon_edge( &edge_rec, &prevnode, face_id );
	    add_polygon_edge(&edge_rec, x, y, &n, &size);
	    if (edge_rec.coord) free(edge_rec.coord);
	}
    }
    MapFace* mf = new MapFace();
    mf->insert_pointers(n, x, y);
    return mf;
}

VpfFeature* VpfFeatureClass::outline_face( row_type row, long int feat_id,
		   long int face_id,
		   vpf_table_type facetable,
		   vpf_table_type ringtable,
		   vpf_table_type edgetable,
		   int ftype,
		   int inner) const {
    face_rec_type face_rec;
    ring_rec_type ring_rec;
    
    face_rec = read_face( face_id, facetable );
    ring_rec = read_ring( face_rec.ring, ringtable );
    MapFace* mf = outline_polygon_loop( row, face_id, ring_rec.edge, edgetable, ftype);
    VpfFeature* vf = new VpfFeature((VpfFeatureClass*)this,feat_id,MapFeature::AreaType,
			  ftype,row);
    vf->face_primitive(mf);
    return vf;
}

VpfFeature* VpfFeatureClass::feature(int i) const {
    return (VpfFeature*)MapFeatureClass::feature(i);
}

/* Inspired by DosVpfviews\'s table_pos */

long int VpfFeatureClass::column_pos(char* colname) const {
   register long int i;
   long int col;
   char altcn[256];

   col = -1;
   for (i = 0; i < _ncolumns; i++) {
       strcpy(altcn, ",:");
       strcat(altcn, colname);
      if (strcasecmp(colname,_columns[i]->name()) == 0 ||
	  strcasecmp(altcn,_columns[i]->name()) == 0) {
	 col = i;
	 break;
      }
   }
   return col;
}

/*****************************************************************************/

VpfColumnHeader::VpfColumnHeader(const header_type header) {
    _hc.name = strdup(header->name);
    strcpy(_hc.description,header->description);
    _hc.keytype = header->keytype;
    strcpy(_hc.vdt,header->vdt);
    _hc.tdx = strdup(header->tdx);
    _hc.type = header->type;
    _hc.count = header->count;
    if (_hc.type == 'T')
	_hc.nullval.Char = strdup(header->nullval.Char);
    else
	_hc.nullval = header->nullval;
    if (header->narrative == NULL)
	_hc.narrative = NULL;
    else
	_hc.narrative = strdup(header->narrative);
}

VpfColumnHeader::VpfColumnHeader(const header_cell header) {
    _hc.name = strdup(header.name);
    strcpy(_hc.description,header.description);
    _hc.keytype = header.keytype;
    strcpy(_hc.vdt,header.vdt);
    _hc.tdx = NULL;
    _hc.type = header.type;
    _hc.count = header.count;
    if (_hc.type == 'T')
	_hc.nullval.Char = strdup(header.nullval.Char);
    else
	_hc.nullval = header.nullval;
    if (header.narrative == NULL)
	_hc.narrative = NULL;
    else
	_hc.narrative = strdup(header.narrative);
}

VpfColumnHeader::~VpfColumnHeader() {
    delete _hc.name;
    delete _hc.tdx;
    if (_hc.type == 'T')
	delete _hc.nullval.Char;
    delete _hc.narrative;
}

VpfColumnHeader* VpfColumnHeader::copy() {
    return new VpfColumnHeader(&_hc);
}

void VpfColumnHeader::print() const {
    printf("Column Name: %s\n", _hc.name);
    printf("Description: %s\n", _hc.description);
    printf("KeyType: %c\n", _hc.keytype);
    printf("Value Description Table name: %s\n", _hc.vdt);
    printf("Type: %c\n", _hc.type);
    printf("Count: %d\n", _hc.count);
    printf("Narrative: %s\n", _hc.narrative);
}

